from youtube_autonomous.database.database_handler import DatabaseHandler
from youtube_autonomous.segments.enums import EnhancementType
from youtube_autonomous.shortcodes.shortcode_parser import ShortcodeParser
from youtube_autonomous.elements.builder.element_builder import ElementBuilder
from youtube_autonomous.elements.rules.effect_element_rules import ElementRules
from youtube_autonomous.elements.rules.rules_checker import RulesChecker
from youtube_autonomous.segments.builder.config import DEFAULT_SEGMENT_PARTS_FOLDER
from yta_multimedia.audio.voice.transcription.stt.whisper import get_transcription_with_timestamps
from yta_general_utils.file import copy_file
from yta_general_utils.file.filename import get_file_extension
from yta_general_utils.temp import get_temp_filename
from yta_general_utils.logger import print_completed
from moviepy.editor import AudioFileClip, VideoFileClip


class Enhancement:
    project_id: int
    segment_index: int
    segment = None
    index: int
    status = ''
    type = ''

    audio_filename: str = ''
    audio_clip = None
    video_filename: str = ''
    video_clip = None
    full_filename: str = ''
    full_clip = None

    audio_narration_filename: str
    voice: str
    narration_text: str

    text: str

    keywords: str

    start: float = None
    duration: float = None

    filename: str
    url: str

    mode: str

    music: str

    enhancements: list

    narration_text_sanitized_without_shortcodes: str = ''
    narration_text_with_simplified_shortcodes: str = ''
    narration_text_sanitized: str = ''
    transcription = None
    shortcodes = None
    created_at = None

    def set_audio_filename(self, audio_filename: str):
        self.audio_filename = audio_filename

        if audio_filename:
            self.audio_clip = AudioFileClip(self.audio_filename)
            self.database_handler.update_project_segment_enhancement_field(self.project_id, self.segment_index, self.index, 'audio_filename', self.audio_filename)

    def set_video_filename(self, video_filename: str):
        self.video_filename = video_filename

        if video_filename:
            self.video_clip = VideoFileClip(self.video_filename)
            self.database_handler.update_project_segment_enhancement_field(self.project_id, self.segment_index, self.index,  'video_filename', self.video_filename)

    def set_full_filename(self, full_filename: str):
        self.full_filename = full_filename

        if full_filename:
            self.full_clip = VideoFileClip(self.full_filename)
            self.database_handler.update_project_segment_enhancement_field(self.project_id, self.segment_index, self.index,  'full_filename', self.full_filename)

    def set_as_finished(self):
        self.database_handler.update_project_segment_enhancement_field(self.project_id, self.segment_index, self.index, 'status', 'finished')

    def __init__(self, project_id, segment_index: int, segment, index: int, data: dict):
        self.project_id = str(project_id)
        self.segment_index = segment_index
        self.segment = segment
        self.index = index

        for key in data:
            setattr(self, key, data[key])

        # I parse shortcodes here because not coming from the 'youtube_autonomous'
        # TODO: Set shortcode tags please (read Notion)
        self.shortcode_parser = ShortcodeParser([])
        if self.narration_text:
            self.shortcode_parser.parse(self.narration_text)
            self.narration_text_sanitized_without_shortcodes = self.shortcode_parser.text_sanitized_without_shortcodes
            self.narration_text_with_simplified_shortcodes = self.shortcode_parser.text_sanitized_with_simplified_shortcodes
            self.narration_text_sanitized = self.shortcode_parser.text_sanitized
            self.shortcodes = self.shortcode_parser.shortcodes

        self.rules = ElementRules.get_subclass_by_type(data['type'])()
        self.builder = ElementBuilder.get_subclass_by_type(data['type'])()
        self.rules_checker = RulesChecker(self.rules)

        self.database_handler = DatabaseHandler()

    # TODO: This below has been moved because inheritance is working strangely

    def create_segment_file(self, filename: str):
        """
        Creates a filename within the definitive segments folder
        to keep the generated file locally to recover it in the
        next project execution if something goes wrong. The 
        definitive filename will be built using the provided
        'filename' and adding some more information in the name.

        This method will generate a temporary filename that uses
        the current segment index in its name and is placed in 
        segment parts folder.

        This method returns the final filename created.
        """
        if not filename:
            raise Exception('No "filename" provided.')

        temp_filename = get_temp_filename(filename)

        return f'{DEFAULT_SEGMENT_PARTS_FOLDER}/segment_{self.index}_{temp_filename}'
    
    def create_narration(self):
        """
        Creates the audio narration (if needed) by generating an AI audio
        narration with provided 'voice' and 'audio_narration_text'
        parameters or by using the 'audio_narration_filename'.

        This method will set the 'audio_filename' to be able to build the
        audio clip in a near future.
        """
        if self.audio_narration_filename:
            segment_part_filename = self.create_segment_file(f'narration.{get_file_extension(self.audio_narration_filename)}')
            copy_file(self.audio_narration_filename, segment_part_filename)
            print_completed('Original voice narration file copied to segment parts folder')
            self.audio_narration_filename = segment_part_filename
            self.set_audio_filename(segment_part_filename)
        else:
            segment_part_filename = self.create_segment_file('narration.wav')
            # TODO: Voice parameter need to change
            self.set_audio_filename(self.builder.build_narration(self.narration_text_sanitized_without_shortcodes, output_filename = segment_part_filename))
            print_completed('Voice narration created successfully')

    def create_transcription(self):
        """
        Creates the transcription of the generated audio narration
        that would be stored in 'self.audio_filename'.
        
        This method returns a words array containing, for each word,
        a 'text', 'start' and 'end' field to be able to use the 
        transcription timestamps.
        """
        transcription = get_transcription_with_timestamps(self.audio_filename, initial_prompt = self.narration_text_sanitized_without_shortcodes)

        # We join all words together as we don't need segments, just words with time
        words = []
        for transcription_segment in transcription['segments']:
            words += transcription_segment['words']
        self.transcription = words

    # TODO: This above has been moved because inheritance is working strangely

    # TODO: This below has to be in the common Object from wich Segment
    # and Enhancement will inherit
    def build_step_1_create_narration(self):
        # 1. Generate narration if needed
        self.create_narration()

    def build_step_2_create_transcription(self):
        # 2. Generate narration transcription
        self.create_transcription()

    def build_step_2_extract_user_shortcodes(self):
        # 2. Extract manually written shortcodes in 'narration_text'
        self.shortcode_parser.parse(self.narration_text)
        self.shortcodes = self.shortcode_parser.shortcodes
        # TODO: Turn shortcodes into enhancements
        enhancements = [shortcode.to_enhancement_element(self.transcription) for shortcode in self.shortcodes]
        print(enhancements)

    def build_step_3_apply_edition_manual_shortcodes(self):
        # 3. Apply shortcodes from Edition Manual and 'narration_text'
        # TODO: By now I'm not applying it
        #self.apply_edition_manual()
        pass

    def build_step_4_build_base_content(self):
        # 2. Build base video
        if not self.video_clip:
            print(self.segment)
            self.video_clip = self.builder.build_from_enhancement(self)
            filename = self.create_segment_file('video.mp4')
            self.video_clip.write_videofile(filename)
            self.set_video_filename(filename)

        # TODO: In the future we could add enhancements to this
        if self.audio_clip:
            self.video_clip = self.video_clip.set_audio(self.audio_clip)
        self.full_clip = self.video_clip
        filename = self.create_segment_file('video.mp4')
        self.full_clip.write_videofile(filename)
        self.set_full_filename(filename)
    
    def build(self):
        """
        Builds this enhancement to be applied in the provided 'segment'.
        """
        if self.type in [EnhancementType.GREENSCREEN.value, EnhancementType.EFFECT.value]:
            self.full_clip = self.builder.build_from_enhancement(self)
        else:
            if self.rules_checker.should_build_narration_rule(self):
                if not self.audio_filename:
                    self.build_step_1_create_narration()

                # We are forcing duration here
                self.duration = self.audio_clip.duration

                if not self.transcription:
                    self.build_step_2_create_transcription()

                # TODO: Handle this (maybe status (?))
                self.build_step_2_extract_user_shortcodes()

                self.build_step_3_apply_edition_manual_shortcodes()

            # TODO: What about duration here that is not set (?)
            if self.audio_clip:
                self.duration = self.audio_clip.duration

            if not self.video_clip:
                print('Building base content step 4')
                self.build_step_4_build_base_content()

            # Make the base video to apply enhancements on it
            if self.audio_clip:
                self.video_clip = self.video_clip.set_audio(self.audio_clip)

            #self.build_step_5_update_enhancements_duration()
            # TODO: By now I'm omitting this part
            #self.build_step_6_build_and_apply_enhancements()

            # TODO: Build the final video
            self.full_clip = self.video_clip
            filename = self.create_segment_file('definitivo.mp4')
            self.full_clip.write_videofile(filename)
            self.set_full_filename(filename)

            self.set_as_finished()

        return self.full_clip


from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from youtube_autonomous.segments.enums import SegmentType, EnhancementType
from youtube_autonomous.elements.validator import BUILDER_SUBCLASSES
from youtube_autonomous.segments.builder.ai import create_ai_narration
from typing import Union


class ElementBuilder:
    # TODO: Maybe __init__ that returns the corresponding subclass
    # according to the provided 'type' (?)
    # https://stackoverflow.com/questions/9187388/possible-to-prevent-init-from-being-called

    @staticmethod
    def get_subclasses():
        from youtube_autonomous.elements.builder.ai_image_element_builder import AIImageElementBuilder
        from youtube_autonomous.elements.builder.image_element_builder import ImageElementBuilder
        from youtube_autonomous.elements.builder.ai_video_element_builder import AIVideoElementBuilder
        from youtube_autonomous.elements.builder.video_element_builder import VideoElementBuilder
        from youtube_autonomous.elements.builder.custom_stock_element_builder import CustomStockElementBuilder
        from youtube_autonomous.elements.builder.stock_element_builder import StockElementBuilder
        from youtube_autonomous.elements.builder.meme_element_builder import MemeElementBuilder
        from youtube_autonomous.elements.builder.sound_element_builder import SoundElementBuilder
        from youtube_autonomous.elements.builder.youtube_video_element_builder import YoutubeVideoElementBuilder
        from youtube_autonomous.elements.builder.text_element_builder import TextElementBuilder
        from youtube_autonomous.elements.builder.premade_element_builder import PremadeElementBuilder
        from youtube_autonomous.elements.builder.effect_element_builder import EffectElementBuilder
        from youtube_autonomous.elements.builder.greenscreen_element_builder import GreenscreenElementBuilder

        return [
            AIImageElementBuilder,
            AIVideoElementBuilder,
            ImageElementBuilder,
            VideoElementBuilder,
            CustomStockElementBuilder,
            StockElementBuilder,
            MemeElementBuilder,
            SoundElementBuilder,
            YoutubeVideoElementBuilder,
            TextElementBuilder,
            PremadeElementBuilder,
            EffectElementBuilder,
            GreenscreenElementBuilder
        ]

    @staticmethod
    def get_subclasses_as_str():
        return ', '.join(ElementBuilder.get_subclasses())
    
    @staticmethod
    def get_subclass_by_type(type: Union[SegmentType, EnhancementType, str]):
        type = ElementParameterValidator.validate_segment_or_enhancement_type(type)

        from youtube_autonomous.elements.builder.ai_image_element_builder import AIImageElementBuilder
        from youtube_autonomous.elements.builder.image_element_builder import ImageElementBuilder
        from youtube_autonomous.elements.builder.ai_video_element_builder import AIVideoElementBuilder
        from youtube_autonomous.elements.builder.video_element_builder import VideoElementBuilder
        from youtube_autonomous.elements.builder.custom_stock_element_builder import CustomStockElementBuilder
        from youtube_autonomous.elements.builder.stock_element_builder import StockElementBuilder
        from youtube_autonomous.elements.builder.meme_element_builder import MemeElementBuilder
        from youtube_autonomous.elements.builder.sound_element_builder import SoundElementBuilder
        from youtube_autonomous.elements.builder.youtube_video_element_builder import YoutubeVideoElementBuilder
        from youtube_autonomous.elements.builder.text_element_builder import TextElementBuilder
        from youtube_autonomous.elements.builder.premade_element_builder import PremadeElementBuilder
        from youtube_autonomous.elements.builder.effect_element_builder import EffectElementBuilder
        from youtube_autonomous.elements.builder.greenscreen_element_builder import GreenscreenElementBuilder

        if type in [SegmentType.MEME, EnhancementType.MEME]:
            return MemeElementBuilder
        elif type in [SegmentType.AI_IMAGE, EnhancementType]:
            return AIImageElementBuilder
        elif type in [SegmentType.AI_VIDEO, EnhancementType.AI_VIDEO]:
            return AIVideoElementBuilder
        elif type in [SegmentType.IMAGE, EnhancementType.IMAGE]:
            return ImageElementBuilder
        elif type in [SegmentType.VIDEO, EnhancementType.VIDEO]:
            return VideoElementBuilder
        elif type in [SegmentType.STOCK, EnhancementType.STOCK]:
            return StockElementBuilder
        elif type in [SegmentType.CUSTOM_STOCK, EnhancementType.CUSTOM_STOCK]:
            return CustomStockElementBuilder
        elif type in [SegmentType.SOUND, EnhancementType.SOUND]:
            return SoundElementBuilder
        elif type in [SegmentType.YOUTUBE_VIDEO, EnhancementType.YOUTUBE_VIDEO]:
            return YoutubeVideoElementBuilder
        elif type in [SegmentType.TEXT, EnhancementType.TEXT]:
            return TextElementBuilder
        elif type in [SegmentType.PREMADE, EnhancementType.PREMADE]:
            return PremadeElementBuilder
        elif type in [EnhancementType.EFFECT]:
            return EffectElementBuilder
        elif type in [EnhancementType.GREENSCREEN]:
            return GreenscreenElementBuilder
    
    @classmethod
    def build_narration(cls, text: str, output_filename: str):
        """
        Generates a narration file that narrates the 'text' provided and
        is stored locally as 'output_filename'. If 'text' or 
        'output_filename' fields are not provided it will raise an 
        Exception.
        """
        ElementParameterValidator.validate_string_mandatory_parameter('text', text)
        ElementParameterValidator.validate_string_mandatory_parameter('output_filename', output_filename)

        return create_ai_narration(text, output_filename = output_filename)
    
    @classmethod
    def handle_narration_from_segment(cls, segment: dict):
        pass


from youtube_autonomous.segments.enums import SegmentField, EnhancementField, SegmentType, EnhancementType, EnhancementMode
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from yta_general_utils.programming.enum import YTAEnum as Enum
from typing import Union



class StringDuration(Enum):
    """
    This string value is only accepted in the segment dict that
    the user provides to the system. The key will be accepted
    and transformed into the value to be processed dynamically
    when building the content.
    """
    SHORTCODE_CONTENT = 99997
    """
    This string value determines that the duration is expected
    to be from the begining of the shortcode content (first 
    word) to the end of it (last word).

    This duration is accepted only in shortcodes.
    """
    FILE_DURATION = 99998
    """
    This string value determines that the duration is expected
    to be the source file (downloaded or obtained from the 
    local system) duration.

    This duration is accepted in shortcodes, in segments and in
    enhancements.
    """
    SEGMENT_DURATION = 99999
    """
    This string value determines that the duration is expected
    to be the whole segment duration (known when built).

    This duration is accepted only in segments and shortcodes.
    """
    
    @staticmethod
    def convert_duration(duration: str):
        """
        Converts the provided 'duration' string to its actual value
        according to the existing StringDuration enums or raises an
        Exception if not valid.

        This method returns the int value of the given 'duration'.
        """
        if duration == StringDuration.SHORTCODE_CONTENT.name:
            duration == StringDuration.SHORTCODE_CONTENT.value
        elif duration == StringDuration.FILE_DURATION.name:
            duration == StringDuration.FILE_DURATION.value
        elif duration == StringDuration.SEGMENT_DURATION.name:
            duration == StringDuration.SEGMENT_DURATION.value
        else:
            raise Exception(f'The provided "duration" parameter {duration} is not a valid StringDuration name.')

        return duration

class ElementValidator:
    """
    Class to validate the segment o enhancement element fields and
    their values.
    """
    SEGMENT_MANDATORY_FIELDS = SegmentField.get_all_values()
    ENHANCEMENT_MANDATORY_FIELDS = EnhancementField.get_all_values()

    @staticmethod
    def validate_segment_fields(segment: dict):
        """
        Validates that the 'segment' dictionary provided has all the
        required segment fields and only those ones and raises an
        Exception if more parameters or less parameters exist.

        This method returns the 'segment' provided if everything is
        ok.
        """
        accepted_fields_str = ', '.join(ElementValidator.SEGMENT_MANDATORY_FIELDS)
        unaccepted_fields = [key for key in segment.keys() if key not in ElementValidator.SEGMENT_MANDATORY_FIELDS]
        unaccepted_fields_str = ', '.join(unaccepted_fields)
        missing_fields = [field for field in ElementValidator.SEGMENT_MANDATORY_FIELDS if field not in segment]
        missing_fields_str = ', '.join(missing_fields)

        # TODO: Improve these 2 messages
        if missing_fields:
            raise Exception(f'The next fields are mandatory and were not found in the Segment: "{missing_fields_str}". The mandatory fields are: "{accepted_fields_str}".')

        if unaccepted_fields:
            raise Exception(f'The next fields are not accepted in Segments by our system: "{unaccepted_fields_str}". The ones accepted are these: "{accepted_fields_str}".')
        
        return segment
    
    @staticmethod
    def validate_enhancement_fields(enhancement: dict):
        """
        Validates that the 'enhancement' dictionary provided has all
        the required segment fields and only those ones and raises an
        Exception if more parameters or less parameters exist.

        This method returns the 'enhancement' provided if everything is
        ok.
        """
        accepted_fields_str = ', '.join(ElementValidator.ENHANCEMENT_MANDATORY_FIELDS)
        unaccepted_fields = [key for key in enhancement.keys() if key not in ElementValidator.ENHANCEMENT_MANDATORY_FIELDS]
        unaccepted_fields_str = ', '.join(unaccepted_fields)
        missing_fields = [field for field in ElementValidator.ENHANCEMENT_MANDATORY_FIELDS if field not in enhancement]
        missing_fields_str = ', '.join(missing_fields)

        # TODO: Improve these 2 messages
        if missing_fields:
            raise Exception(f'The next fields are mandatory and were not found in the Enhancement: "{missing_fields_str}". The mandatory fields are: "{accepted_fields_str}".')

        if unaccepted_fields:
            raise Exception(f'The next fields are not accepted in Enhancement by our system: "{unaccepted_fields_str}". The ones accepted are these: "{accepted_fields_str}".')
        
        return enhancement
    
    @staticmethod
    def validate_segment_duration_field(duration: Union[int, float, str, None]):
        """
        Validates that the 'duration' provided, if not None, has a 
        valid and positive numeric value or is a string accepted 
        for a segment.
        """
        if duration is not None:
            if isinstance(duration, str):
                accepted_str_duration = [StringDuration.FILE_DURATION.name]
                accepted_str_duration_str = ', '.join(accepted_str_duration)
                if duration not in accepted_str_duration:
                    raise Exception(f'Unexpected duration value {duration}. The accepted duration strings for a segment are: {accepted_str_duration_str}')

                duration = StringDuration[duration]
            else:
                ElementParameterValidator.validate_positive_number('duration', duration)

        return duration

    @staticmethod
    def validate_enhancement_duration_field(duration: Union[int, float, str, None]):
        """
        Validates that the 'duration' provided, if not None, has a 
        valid and positive numeric value or is a string accepted 
        for an enhancement.

        This method will return the duration, if string, as a
        StringDuration Enum Object, or if a number, as it is.
        """
        if duration is not None:
            if isinstance(duration, str):
                accepted_str_duration = [StringDuration.FILE_DURATION.name, StringDuration.SEGMENT_DURATION.name]
                accepted_str_duration_str = ', '.join(accepted_str_duration)
                if duration not in accepted_str_duration:
                    raise Exception(f'Unexpected duration value {duration}. The accepted duration strings for an enhancement are: {accepted_str_duration_str}')
                
                duration = StringDuration[duration]
            else:
                ElementParameterValidator.validate_positive_number('duration', duration)
        
        return duration
    
    @staticmethod
    def validate_enhancement_mode_field(mode: Union[EnhancementMode, str, None]):
        """
        Validates if the provided 'mode' parameter is a valid parameter
        for an Enhancement object. It should be an EnhancementMode enum
        or one of its string values to be accepted.
        
        This method will raise an Exception if not valid or will return
        it as it is, or as an EnhancementMode if one of its valid string
        values provided.
        """
        if mode is not None:
            if isinstance(mode, str):
                if not EnhancementMode.is_valid(mode):
                    raise Exception(f'Unexpected mode value {mode}. The accepted mode strings for an enhancement are: {EnhancementMode.get_all_values_as_str()}')
                
                mode = EnhancementMode(mode)
            elif not isinstance(mode, EnhancementMode):
                raise Exception('The "mode" parameter provided is not a EnhancementMode nor a string.')
        
        return mode
    
    @staticmethod
    def validate_enhancement_mode_field_for_type(mode: Union[str, None], type: str):
        """
        Validates if the provided 'mode' is valid for the also provided
        'type', raising an Exception if not or returning the mode as it
        is if valid.
        """
        # TODO: Refactor this. These mode valid if type and that stuff should
        # be centralized in one file or using another strategy, but for now
        # it is like this due to issues in the building process
        INLINE_VALID_TYPES = [EnhancementType.AI_IMAGE.value, EnhancementType.IMAGE.value, EnhancementType.AI_VIDEO.value, EnhancementType.VIDEO.value, EnhancementType.STOCK.value, EnhancementType.CUSTOM_STOCK.value, EnhancementType.MEME.value, EnhancementType.SOUND.value, EnhancementType.PREMADE.value, EnhancementType.YOUTUBE_VIDEO.value, EnhancementType.TEXT.value]
        OVERLAY_VALID_TYPES = [EnhancementType.AI_IMAGE.value, EnhancementType.IMAGE.value, EnhancementType.AI_VIDEO.value, EnhancementType.VIDEO.value, EnhancementType.STOCK.value, EnhancementType.CUSTOM_STOCK.value, EnhancementType.MEME.value, EnhancementType.SOUND.value, EnhancementType.PREMADE.value, EnhancementType.YOUTUBE_VIDEO.value, EnhancementType.TEXT.value]
        REPLACE_VALID_TYPES = [EnhancementType.EFFECT.value, EnhancementType.GREENSCREEN.value]

        if (mode == EnhancementMode.INLINE.value and type not in INLINE_VALID_TYPES) or (mode == EnhancementMode.OVERLAY.value and type not in OVERLAY_VALID_TYPES) or (mode == EnhancementMode.REPLACE.value and type not in REPLACE_VALID_TYPES):
            raise Exception(f'The provided "mode" parameter {mode} is not valid for the also provided "type" parameter {type}.')
        
        return mode
    
    @staticmethod
    def validate_segment_duration_field_for_type(duration: Union[int, float, str, None], type: str):
        """
        Validates if the provided 'duration' is a valid duration for the
        also provided Segment 'type'. The 'FILE_DURATION' duration is 
        only accepted for the 'meme' and 'sound' type'.
        """
        if duration in [StringDuration.FILE_DURATION, StringDuration.FILE_DURATION.name] and type not in [SegmentType.MEME.value, SegmentType.SOUND.value]:
            raise Exception(f'The "{StringDuration.FILE_DURATION.name}" string duration can be set only as {SegmentType.MEME.value} and {SegmentType.SOUND.value} for Segment types.')
        
    @staticmethod
    def validate_enhancement_duration_field_for_type(duration: Union[int, float, str, None], type: str):
        """
        Validates if the provided 'duration' is a valid duration for the
        also provided Enhancement 'type'. The 'FILE_DURATION' duration is
        only accepted for the 'meme' and 'sound' type'.
        """
        if duration in [StringDuration.FILE_DURATION, StringDuration.FILE_DURATION.name] and type not in [EnhancementType.MEME.value, EnhancementType.SOUND.value]:
            raise Exception(f'The "{StringDuration.FILE_DURATION.name}" string duration can be set only as {EnhancementType.MEME.value} and {EnhancementType.SOUND.value} for Enhancement types.')
    
    @staticmethod
    def validate_shortcode_duration_field(duration: Union[int, float, str, None]):
        """
        Validates that the 'duration' provided, if not None, has a 
        valid and positive numeric value or is a string accepted 
        for a shortcode.
        """
        if duration is not None:
            if isinstance(duration, str):
                accepted_str_duration = [StringDuration.FILE_DURATION.name, StringDuration.SEGMENT_DURATION.name, StringDuration.SHORTCODE_CONTENT]
                accepted_str_duration_str = ', '.join(accepted_str_duration)
                if duration not in accepted_str_duration:
                    raise Exception(f'Unexpected duration value {duration}. The accepted duration strings for a shortcode are: {accepted_str_duration_str}')
            else:
                ElementParameterValidator.validate_positive_number('duration', duration)
        
        return duration


from youtube_autonomous.database.database_handler import DatabaseHandler
from youtube_autonomous.segments.validation.segment_validator import SegmentValidator
from youtube_autonomous.shortcodes.objects.shortcode_tag import ShortcodeTag, ShortcodeType
from youtube_autonomous.elements.validator.element_validator import ElementValidator
from youtube_autonomous.elements.rules.element_rules import ElementRules
from youtube_autonomous.project.project import Project
from youtube_autonomous.elements.validator.element_validator import StringDuration
from youtube_autonomous.elements.rules.rules_checker import RulesChecker
from youtube_autonomous.segments.enums import SegmentField, EnhancementField, ProjectStatus, SegmentStatus, SegmentType, EnhancementType
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from youtube_autonomous.segments.builder.config import DEFAULT_SEGMENT_PARTS_FOLDER, DEFAULT_PROJECTS_OUTPUT_FOLDER
from youtube_autonomous.shortcodes.shortcode_parser import ShortcodeParser
from yta_general_utils.temp import clean_temp_folder
from yta_general_utils.programming.path import get_project_abspath
from yta_general_utils.path import create_file_abspath
from yta_general_utils.file.reader import read_json_from_file
from yta_general_utils.logger import print_completed, print_in_progress, print_error
from datetime import datetime
from typing import Union

import copy


class YoutubeAutonomous:
    __database_handler: DatabaseHandler = None
    """
    Object to interact with the database and get and create projects.

    _This parameter is not manually set by the user._
    """
    __segment_validator: SegmentValidator = None
    """
    Object to validate the segments we want to use to build a project.

    _This parameter is not manually set by the user._
    """
    __segments_abspath: Union[str, None] = None
    """
    The absolute path that points to the segments folder in which this
    software will store the segments related content parts that are
    created during the whole segment building process. This folder 
    allows us recovering a non-finished project in which we generated
    some segments in the past but not all of them.

    _This parameter is not manually set by the user._
    """
    __projects_output_abspath: Union[str, None] = None
    """
    The absolute path that points to the projects output folder in which
    this software will store the final videos generated.

    _This parameter is not manually set by the user._
    """

    def __init__(self):
        """
        Initializes the object and creates the segment part building
        folder in which segment parts will be stored.
        """
        self.__database_handler = DatabaseHandler()
        self.__segment_validator = SegmentValidator()

        self.__segments_abspath = get_project_abspath() + DEFAULT_SEGMENT_PARTS_FOLDER + '/'
        self.__projects_output_abspath = get_project_abspath() + DEFAULT_PROJECTS_OUTPUT_FOLDER + '/'

        # We force to create the folder if it doesn't exist
        create_file_abspath(self.__segments_abspath + 'toforce')
        create_file_abspath(self.__projects_output_abspath + 'toforce')

    def purge(self, do_remove_segments_files = False):
        """
        Cleans the temporary folder removing all previous generated 
        temporary files, and also the segment files if the
        'do_remove_segments_files' parameter is set to True.
        """
        clean_temp_folder()
        if do_remove_segments_files:
            # TODO: Remove all files in self.__segments_abspath folder
            pass

    def check_config(self):
        # TODO: Check that he config is ok
        pass
     
    def insert_project_in_database_from_file(self, filename: str):
        """
        Reads the provided project content 'filename' and creates a new 
        project in the database if the provided 'filename' contains a new
        project and is valid. If the information belongs to an already
        registered project, it will raise an exception.

        This method returns the new stored project mongo ObjectId if 
        successfully stored, or raises an Exception if anything went wrong.
        """
        ElementParameterValidator.validate_filename(filename, None)
        
        json_data = self.__read_project_from_file(filename)

        # If a project with the same content exists, it is the same project
        db_project = self.__database_handler.get_database_project_from_json(json_data)
        if db_project:
            raise Exception('There is an existing project in the database with the same content.')

        # Process the information to store it parsed
        json_data = self.validate_and_prepare_data_for_database(json_data)

        db_project = self.__database_handler.insert_project(json_data)

        print_completed('Project created in database with ObjectId = "' + str(db_project['_id']) + '"')

        return str(db_project['_id'])
    
    def validate_data(self, json_data: dict):
        segments = json_data.get('segments', [])

        if not segments:
            raise Exception('No "segments" in the provided "json_data" parameter.')

        # TODO: Check main structure and validate fields
        # TODO: This shortcode parser must have the definitive shortcodes
        # TODO: Shortcode handlers must be working, check Notion (https://www.notion.so/ShortcodeParser-personalizado-10ff5a32d462804d9253e9cb782e5540?pvs=4)
        shortcode_parser = ShortcodeParser([ShortcodeTag('meme', ShortcodeType.SIMPLE)])

        for index, segment in enumerate(segments):
            print_in_progress(f'Handling segment [{index}]')

            # Check all fields are set (no more, no less)
            ElementValidator.validate_segment_fields(segment)

            type = ElementParameterValidator.validate_segment_type(segment.get(SegmentField.TYPE.value, '')).value

            # Check if 'text' has any shortcode
            try:
                ShortcodeParser([]).parse(segment.get(SegmentField.TEXT.value, ''))
            except Exception:
                print_error(f'Field "{SegmentField.TEXT.value}" not found or shortcodes found on it.')
                exit()

            # Check if 'narration_text' has unaccepted shortcodes
            try:
                shortcode_parser.parse(segment.get(SegmentField.NARRATION_TEXT.value, ''))
            except Exception:
                print_error(f'Field "{SegmentField.NARRATION_TEXT.value}" not found or unregistered shortcodes found on it.')
                exit()

            # Check 'duration' is valid segment duration string or 
            # positive numeric value
            duration = ElementValidator.validate_segment_duration_field(segment.get(SegmentField.DURATION.value, None))

            # We can only accept 'FILE_DURATION' for some specific types
            ElementValidator.validate_segment_duration_field_for_type(duration, type)
            
            for index, enhancement in enumerate(segment.get(SegmentField.ENHANCEMENTS.value, [])):
                print_in_progress(f'   Handling enhancement [{index}]')

                # We let the user know if any unaccepted field has been given
                # and we raise and Exception if that happen
                ElementValidator.validate_enhancement_fields(enhancement)

                # Check enhancement type is allowed as enhancement
                enhancement_type = ElementParameterValidator.validate_enhancement_type(enhancement.get(EnhancementField.TYPE.value, '')).value
                
                # Check if 'text' has any shortcode
                try:
                    ShortcodeParser([]).parse(enhancement.get(EnhancementField.TEXT.value, ''))
                except Exception:
                    print_error(f'Field "{EnhancementField.TEXT.value}" not found or shortcodes found on it.')
                    exit()

                # Check if 'narration_text' has unaccepted shortcodes
                try:
                    shortcode_parser.parse(enhancement.get(EnhancementField.NARRATION_TEXT.value, ''))
                except Exception:
                    print_error(f'Field "{EnhancementField.NARRATION_TEXT.value}" not found or unregistered shortcodes found on it.')
                    exit()

                # Check 'duration' is valid enhanacement duration string or
                # positive numeric value
                duration = ElementValidator.validate_enhancement_duration_field(enhancement.get(EnhancementField.DURATION.value, None))
                mode = ElementValidator.validate_enhancement_mode_field(enhancement.get(EnhancementField.MODE.value, None)).value

                # We can only accept 'FILE_DURATION' for some specific types
                ElementValidator.validate_enhancement_duration_field_for_type(duration, type)

                # We can only accept the mode if its type accepts it
                ElementValidator.validate_enhancement_mode_field_for_type(mode, type)

                RulesChecker.check_need_rules(enhancement, ElementRules.get_subclass_by_type(enhancement_type)())
                # TODO: We should check any other thing that is here (frames, etc.)

            # 4. If doesn't have necessary fields for its type
            RulesChecker.check_need_rules(segment, ElementRules.get_subclass_by_type(type)())
    
    def prepare_data_for_database(self, json_data: dict):
        """
        This method prepares the project data that has been previously
        verified so it has the structure with it will be stored in the
        database.

        This method will parse the 'narration_text' and generate the
        some processed and sanitized fields (it processes the 
        shortcodes). It will also transform any string type duration
        to its actual numeric value that will be dynamically processed
        later when building the content.

        The result of this method will be a dict containing 'status',
        'script' and 'segments' fields.
        """
        # This is the structure the project must have in database
        data = {
            'status': ProjectStatus.TO_START.value,
            'script': copy.deepcopy(json_data),
            'segments': None
        }

        shortcode_parser = ShortcodeParser([])
        segments = []
        for segment in json_data['segments']:
            # Manually handle 'narration_text'
            if segment.get(SegmentField.NARRATION_TEXT.value, ''):
                shortcode_parser.parse(segment[SegmentField.NARRATION_TEXT.value])
                segment['narration_text_sanitized_without_shortcodes'] = shortcode_parser.text_sanitized_without_shortcodes
                segment['narration_text_with_simplified_shortcodes'] = shortcode_parser.text_sanitized_with_simplified_shortcodes
                segment['narration_text_sanitized'] = shortcode_parser.text_sanitized

            # Manually handle string duration
            if segment.get(SegmentField.DURATION.value, None) and isinstance(segment[SegmentField.DURATION.value], str):
                segment[SegmentField.DURATION.value] = StringDuration.convert_duration(segment[SegmentField.DURATION.value])
                    
            segment['status'] = SegmentStatus.TO_START.value
            segment['created_at'] = datetime.now()

            for enhancement in segment.get(SegmentField.ENHANCEMENTS.value, []):
                # Manually handle 'narration_text'
                if enhancement.get(EnhancementField.NARRATION_TEXT.value, ''):
                    shortcode_parser.parse(enhancement[EnhancementField.NARRATION_TEXT.value])
                    enhancement['narration_text_sanitized_without_shortcodes'] = shortcode_parser.text_sanitized_without_shortcodes
                    enhancement['narration_text_with_simplified_shortcodes'] = shortcode_parser.text_sanitized_with_simplified_shortcodes
                    enhancement['narration_text_sanitized'] = shortcode_parser.text_sanitized

                # Manually handle string duration
                if enhancement.get('duration', None) and isinstance(enhancement['duration'], str):
                    enhancement['duration'] = StringDuration.convert_duration(enhancement['duration'])

                enhancement['status'] = SegmentStatus.TO_START.value
                enhancement['created_at'] = datetime.now()

            segments.append(segment)

        data['segments'] = segments

        return data

    def validate_and_prepare_data_for_database(self, json_data: dict):
        """
        This method will validate the 'json_data' provided and raise
        an Exception if anything is wrong. If it is ok, it will 
        process the data and turn it into the expected data to be 
        stored in the database.
        """
        self.validate_data(json_data)
        return self.prepare_data_for_database(json_data)
        
    
    def process_unfinished_projects(self):
        db_projects = self.get_unfinished_projects()

        for db_project in db_projects:
            print_in_progress(f'Processing project "{db_project["_id"]}"')
            self.process_project(db_project)
            print_completed(f'Project "{db_project["_id"]}" processed succesfully!')

    def process_project(self, db_project):
        project = Project(db_project['_id'])
        project.build(f'{self.__projects_output_abspath}project_{project.id}.mp4')
    
    # TODO: This below is for testing, remove it
    def get_unfinished_project(self):
        """
        Returns the first unfinished project that exists in the
        database, or None if there are no unfinished projects.
        """
        return self.__database_handler.get_unfinished_project()

    def get_unfinished_projects(self):
        """
        Returns a list containing all the existing projects in the 
        database that have not been finished yet, or an empty list
        if there are no unfinished projects.
        """
        return self.__database_handler.get_unfinished_projects()
    
    def __read_project_from_file(self, filename: str):
        """
        Reads the provided 'filename', that should contain the information of
        a project, and validates its content and structure. This method will
        return the data read as a dict or will raise an Exception if something
        is not ok.
        """
        if not filename:
            raise Exception('No "filename" provided.')
        
        if not isinstance(filename, str):
            raise Exception('The "filename" parameter provided is a str.')
        
        json_data = read_json_from_file(filename)

        if not 'segments' in json_data:
            raise Exception('The provided "filename" does not contain the expected data structure.')

        # We validate each segment to be able to store the project
        # This will raise an Exception if something is not ok
        # for segment in json_data['segments']:
        #     self.__segment_validator.validate(segment)

        return json_data

    # TODO: Apply ObjectId type to 'project_id'
    def __get_project_from_database_by_id(self, project_id):
        """
        Gets the project with the provided 'project_id' if existing, or
        None if not.
        """
        return self.__database_handler.get_database_project_from_id(project_id)
    
    # TODO: Apply json type to 'json'
    def __get_project_from_database_by_json(self, json):
        """
        Gets the project with the provided 'json' if existing, or None if
        not. This 'json' will be compared with the 'script' field stored
        in database to check if there is a similar project previously 
        stored that must be recovered instead of duplicating it.
        """
        return self.__database_handler.get_database_project_from_json(json)



from yta_general_utils.programming.enum import YTAEnum as Enum


class ShortcodeType(Enum):
    """
    This Enum represents the different type of shortcodes that
    we can handle according to their scopes. It could be simple
    [tag] shortcode or a block-scoped one [tag] ... [/tag].
    """
    BLOCK = 'block'
    SIMPLE = 'simple'


from youtube_autonomous.elements.builder.element_builder import ElementBuilder
from youtube_autonomous.segments.enums import SegmentType
from youtube_autonomous.segments.enums import SegmentField, EnhancementField
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from yta_multimedia.video.generation.google.google_search import GoogleSearch
from yta_multimedia.video.generation.google.youtube_search import YoutubeSearch
from typing import Union


class PremadeElementBuilder(ElementBuilder):
    @classmethod
    def build_from_enhancement(cls, enhancement: dict):
        premade_name = enhancement.get(EnhancementField.KEYWORDS.value, '')

        ElementParameterValidator.validate_premade_name(premade_name)

        return cls.build_custom_from_premade_name(premade_name)

    @classmethod
    def build_from_segment(cls, segment: dict):
        premade_name = segment.get(SegmentField.KEYWORDS.value, '')
        ElementParameterValidator.validate_premade_name(premade_name)

        # TODO: Obtain the other parameters

        return cls.build_custom_from_premade_name(premade_name)

    @classmethod
    def premade_name_to_class(cls, premade_name: str):
        """
        Returns the corresponding premade class according to the
        provided 'premade_name'. If no premade class found, the
        return will be None.
        """
        premade_class = None

        if premade_name == SegmentType.GOOGLE_SEARCH.value:
            premade_class = GoogleSearch
        elif premade_name == SegmentType.YOUTUBE_SEARCH.value:
            premade_class = YoutubeSearch
        
        return premade_class

    @classmethod
    def build_custom_from_premade_name(cls, premade_name: str, **parameters):
        ElementParameterValidator.validate_premade_name(premade_name)

        premade_class = cls.premade_name_to_class(premade_name)
        if not premade_class:
            raise Exception(f'The provided "premade_name" parameter {premade_name} is not a valid premade name.')

        return cls.build_custom(premade_class, **parameters)

    @classmethod
    def build_custom(cls, premade, **parameters):
        # TODO: Make the premades implement an abstract class named
        # 'Premade' to be able to detect them as subclasses
        return premade(**parameters).generate()

    @classmethod
    def build(cls, text: str, duration: Union[float, int]):
        """
        Basic example to test that the building process and
        the class are working correctly.

        TODO: Remove this in the future when 'custom' is 
        working perfectly.
        """
        ElementParameterValidator.validate_text(text)
        ElementParameterValidator.validate_duration(duration)

        return GoogleSearch(text).generate()